internal_function
B32 storageMakeDirectoryIfNotExist (Char *path, Char **error)
{
    struct stat st = {0};

    if(stat(path, &st) == -1) {
        if(mkdir(path, 0777) < 0) {
            *error = strdup("IO Error: Can not create data container");
            return false;
        }
    }

    return true;
}

internal_function
B32 storageSetFileData(U64 hash, Char *key, Char *value, Char **error)
{
    Char byte_one_two[5] = {0};
    Char byte_three_four[5] = {0};
    Char byte_five_six[5] = {0};
    Char byte_seven_eight[5] = {0};

    Char path[300] = {0};

    FILE *file = NULL;

    snprintf(byte_one_two, 5, "%04hx", (U16)((hash & 0xFFFF000000000000) >> 48));
    snprintf(byte_three_four, 5, "%04hx", (U16)((hash & 0xFFFF00000000) >> 32));
    snprintf(byte_five_six, 5, "%04hx", (U16)((hash & 0xFFFF0000) >> 16));
    snprintf(byte_seven_eight, 5, "%04hx",(U16)(hash & 0xFFFF));

    /* printf("SET: %s %s %s %s\n\t%s\n\t\t%s\n", */
    /*        byte_one_two, byte_three_four, byte_five_six, byte_seven_eight, */
    /*        key, value); */

    strcat(path, byte_one_two);
    strcat(path, "/");
    if (!storageMakeDirectoryIfNotExist(path, error)) goto error;

    strcat(path, byte_three_four);
    strcat(path, "/");
    if (!storageMakeDirectoryIfNotExist(path, error)) goto error;

    strcat(path, byte_five_six);
    strcat(path, "/");
    if (!storageMakeDirectoryIfNotExist(path, error)) goto error;

    strcat(path, byte_seven_eight);
    strcat(path, "/");
    if (!storageMakeDirectoryIfNotExist(path, error)) goto error;

    strcat(path, key);
    if((file = fopen(path, "w")) != NULL) {
        fprintf(file, "%s", value);
        fclose(file);
    } else {
        *error = strdup("IO Error: Cannot store data");
        goto error;
    }

    cacheSetKV(hash, key, value);

    return true;

error:
    return false;
}

internal_function
Char* storageGetFileData(U64 hash, Char *key, Char **error)
{
    Char *cached_value = cacheGetKV(hash, key, NULL);
    if (cached_value != NULL) {
        return strdup(cached_value);
    }

    Char byte_one_two[5] = {0};
    Char byte_three_four[5] = {0};
    Char byte_five_six[5] = {0};
    Char byte_seven_eight[5] = {0};

    Char path[300] = {0};

    Char *value = calloc((KiB(256) + 1), sizeof(*value));
    FILE *file = NULL;

    snprintf(byte_one_two, 5, "%04hx", (U16)((hash & 0xFFFF000000000000) >> 48));
    snprintf(byte_three_four, 5, "%04hx", (U16)((hash & 0xFFFF00000000) >> 32));
    snprintf(byte_five_six, 5, "%04hx", (U16)((hash & 0xFFFF0000) >> 16));
    snprintf(byte_seven_eight, 5, "%04hx",(U16)(hash & 0xFFFF));

    /* printf("GET: %s %s %s %s\n\t%s\n", */
    /*        byte_one_two, byte_three_four, byte_five_six, byte_seven_eight, */
    /*        key); */

    strcat(path, byte_one_two);
    strcat(path, "/");
    strcat(path, byte_three_four);
    strcat(path, "/");
    strcat(path, byte_five_six);
    strcat(path, "/");
    strcat(path, byte_seven_eight);
    strcat(path, "/");
    strcat(path, key);

    if(access(path, F_OK) == 0) {
        if((file = fopen(path, "r")) != NULL) {
            fseek(file, 0L, SEEK_END);
            Size size = ftell(file);
            fseek(file, 0L, SEEK_SET);
            fread(value, size, sizeof(Char), file);
            fclose(file);
        } else {
            *error = strdup("IO Error: Cannot retrieve data");
         goto error;
       }
    } else {
        *error = strdup("Does not exist");
         goto error;
    }

    return value;

 error:
    free(value);
    return NULL;
}

/* internal_function */
/* void storageConvertPathToParent (Char *str) */
/* { */
/*     Size s = strlen(str); */

/*     while ((str[s] != '/') && (s != 0)) { */
/*         str[s] = '\0'; */
/*         s--; */
/*     } */

/*     str[s] = '\0'; */

/*     if (s == 0) { */
/*         str[0] = '.'; */
/*     } */

/*     return; */
/* } */

/* internal_function */
/* B32 storageDeleteEmptyDirectoryAndSwitchToParent (Char *path) */
/* { */
/*     B32 result = false; */

/*     struct dirent *d; */
/*     int n = 0; */
/*     DIR *dir = opendir(path); */
/*     if (dir != NULL) { */
/*         while ((d = readdir(dir)) != NULL) { */
/*             if(++n > 2) { */
/*                 break; */
/*             } */
/*         } */

/*         closedir(dir); */

/*         if (n <= 2) { // Directory Empty */
/*             rmdir(path); */
/*             result = true; */
/*         } */
/*     } */

/*     storageConvertPathToParent(path); */

/*     return result; */
/* } */

internal_function
B32 storageDeleteFileData(U64 hash, Char *key, Char **error)
{
    Char byte_one_two[5] = {0};
    Char byte_three_four[5] = {0};
    Char byte_five_six[5] = {0};
    Char byte_seven_eight[5] = {0};

    Char path[300] = {0};

    snprintf(byte_one_two, 5, "%04hx", (U16)((hash & 0xFFFF000000000000) >> 48));
    snprintf(byte_three_four, 5, "%04hx", (U16)((hash & 0xFFFF00000000) >> 32));
    snprintf(byte_five_six, 5, "%04hx", (U16)((hash & 0xFFFF0000) >> 16));
    snprintf(byte_seven_eight, 5, "%04hx",(U16)(hash & 0xFFFF));

    /* printf("DEL: %s %s %s %s\n\t%s\n", */
    /*        byte_one_two, byte_three_four, byte_five_six, byte_seven_eight, */
    /*     key); */

    strcat(path, byte_one_two);
    strcat(path, "/");
    strcat(path, byte_three_four);
    strcat(path, "/");
    strcat(path, byte_five_six);
    strcat(path, "/");
    strcat(path, byte_seven_eight);
    strcat(path, "/");
    strcat(path, key);

    // FIXME(namna): This is in a race condition with the SetData
    int ret = 0;
    if(access(path, F_OK) == 0) {
        if((ret = remove(path)) != 0) {
            *error = strdup("IO Error: Cannot delete data");
            goto error;
        }
    } else {
        *error = strdup("Does not exist");
         goto error;
    }

    /* storageConvertPathToParent(path); // Remove file name */
    /* if (storageDeleteEmptyDirectoryAndSwitchToParent(path)) { */
    /*     if (storageDeleteEmptyDirectoryAndSwitchToParent(path)) { */
    /*         if (storageDeleteEmptyDirectoryAndSwitchToParent(path)) { */
    /*             storageDeleteEmptyDirectoryAndSwitchToParent(path); */
    /*         } */
    /*     } */
    /* } */

    Size cache_slot = 0;
    Char *cached_value = cacheGetKV(hash, key, &cache_slot);
    if (cached_value != NULL) {
        cacheInvalidateSlot(cacheGetBucket(hash), cache_slot);
    }

    return true;

 error:
    return false;
}
